# automatically generated by the FlatBuffers compiler, do not modify

# namespace: tosa

class DType(object):
    UNKNOWN = 0
    BOOL = 1
    UINT8 = 2
    INT4 = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    INT48 = 7
    FLOAT = 8

